unit Logger;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  StdCtrls,
  Buttons,
  ComCtrls,
  ExtCtrls;

type
  { Exceptions can happen for expected reasons, such as for numeric conversion overflows and the like. }
  TLogEntryType = (leInformation, leUnknown, leAssertion, leException, leError);
  pLogEntry = ^TLogEntry;
  TLogEntry = record
    TimeOfEntry: string[20];
    LogType: TLogEntryType;
    ModuleName: string[20];
    ProcedureName: string[20];
    LineNo: integer;
    MessageText: string[40];
  end;

  TFormLogger = class(TForm)
    BitBtn1: TBitBtn;
    BitBtnRefresh: TBitBtn;
    TimerLogWrite: TTimer;
    RichEditLogView: TRichEdit;
    procedure FormCreate(Sender: TObject);
    procedure BitBtnRefreshClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TimerLogWriteTimer(Sender: TObject);
  private
    FLogRecords: TList;
    FIgnoreLoggingErrors: boolean;
    FLogFilePath: string;
    FLogFile: TextFile;
    procedure WriteLogEntryToFile(aLogEntry: PLogEntry);
    procedure SetLogFilePath(newPath: string);
    procedure RefreshViewer;
    function FormatEntry(aLogEntry: PLogEntry): string;
    procedure WriteLog;

    { Private declarations }
  public
    procedure Log(alogType: TLogEntryType; aModuleName, aProcedureName, aMessage: string; alineNo: integer);
    property LogFilePath: string read FLogFilePath write SetLogFilePath;
    procedure CloseLog;
  end;

  { Global access to logging }
function Log(aMessage: string; alogType: TLogEntryType; aModuleName, aProcedureName: string; alineNo: integer): boolean;

var
  FormLogger: TFormLogger;
  LogEntryType: TLogEntryType;

implementation

{$R *.DFM}

function Log(aMessage: string; alogType: TLogEntryType; aModuleName, aProcedureName: string; alineNo: integer): boolean;
begin
  Result:=False;
  if Assigned(FormLogger) then
  Begin
    FormLogger.Log(alogType, aModuleName, aProcedureName, aMessage, alineNo);
    Result:=True;
  End;
end;

procedure TFormLogger.SetLogFilePath(newPath: string);
begin

end;

procedure TFormLogger.Log(alogType: TLogEntryType; aModuleName, aProcedureName, aMessage: string; alineNo: integer);
var
  logEntry: PLogEntry;
begin

  New(logEntry);

  logEntry^.TimeOfEntry := FormatDateTime('yymmdd,hh:nn:ss:zzz', Now);
  logEntry^.LogType := alogType;
  logEntry^.ModuleName := aModuleName;
  logEntry^.ProcedureName := aProcedureName;
  logEntry^.MessageText := aMessage;
  logEntry^.LineNo := alineNo;

  FLogRecords.Add(logEntry); // Add to log list

end;

procedure TFormLogger.WriteLog;
var
  i: integer;
begin

  try
    for i := 0 to FLogRecords.Count - 1 do begin
      WriteLogEntryToFile(FLogRecords[i]);
    end;
    FLogRecords.Clear; // the log for new entries

  except
    on E: Exception do
      if not FIgnoreLoggingErrors then
        if MessageDlg('Could not Log error due to [' + E.message + '] should we abort PNCS?', mtError, [mbYes, mbNo], 0) = mrYes then
          halt // Scram
        else begin
          FIgnoreLoggingErrors := true;
          MessageDlg('Logging is no longer being saved, you should restart PNCS at your earliest convienence.', mtWarning, [mbOK], 0);
        end;
  end;
end;

procedure TFormLogger.WriteLogEntryToFile(aLogEntry: PLogEntry);
begin

  if not FileExists(FLogFilePath) then begin
    AssignFile(FLogFile, FLogFilePath);
    Rewrite(FLogFile);
  end;

  WriteLn(FLogFile, FormatEntry(aLogEntry));
  Flush(FLogFile);

end;

function TFormLogger.FormatEntry(aLogEntry: PLogEntry): string;
var
  TimeStr, LogTypeStr, ModuleNameStr, ProcedureNameStr, MessageTextStr, LineNoStr: string;
begin
  with aLogEntry^ do begin
    TimeStr := TimeOfEntry;
    MessageTextStr := MessageText;
    case LogType of
      leInformation: LogTypeStr := 'INFORM';
      leUnknown: LogTypeStr := 'UNKNOWN';
      leAssertion: LogTypeStr := 'ASSERT';
      leException: LogTypeStr := 'EXCEPTION';
      leError: LogTypeStr := 'ERROR';
      else
        LogTypeStr := 'UNKNOWN';
    end;
    ModuleNameStr := ModuleName;
    ProcedureNameStr := ProcedureName;
    try
      LineNoStr := IntToStr(LineNo);
    except
    end;
  end;
  Result := TimeStr + ',' +
    LogTypeStr + ',' +
    MessageTextStr + ',' +
    ModuleNameStr + '.' + ProcedureNameStr + '[' + LineNoStr + ']';
end;

procedure TFormLogger.FormCreate(Sender: TObject);
begin
  FIgnoreLoggingErrors := false;
  FLogFilePath := GetCurrentDir + '\' + 'PNCSLog_' + FormatDateTime('yymmddhhnnss', Now) + '.log';
  FLogRecords := TList.Create;
end;

procedure TFormLogger.BitBtnRefreshClick(Sender: TObject);
begin
  RefreshViewer;
end;

procedure TFormLogger.FormShow(Sender: TObject);
begin
  RefreshViewer;
end;

procedure TFormLogger.RefreshViewer;
begin
  CloseLog;
  RichEditLogView.Lines.Clear;
  RichEditLogView.Lines.LoadFromFile(FLogFilePath);
  Append(FLogFile);
end;

procedure TFormLogger.CloseLog;
begin
  WriteLog; // any remaining rows
  CloseFile(FLogFile); // Auto Flush then close file
end;

procedure TFormLogger.FormDestroy(Sender: TObject);
begin
  CloseLog;
  FLogRecords.Free;
end;

procedure TFormLogger.TimerLogWriteTimer(Sender: TObject);
begin
  try
    TimerLogWrite.Enabled := false;
    WriteLog; // write the entries (if any)
  finally
    TimerLogWrite.Enabled := true;
  end;
end;

end.
