library VACAPRIVVA;

{ Important note about DLL memory management: ShareMem must be the
  first unit in your library's USES clause AND your project's (select
  Project-View Source) USES clause if your DLL exports any procedures or
  functions that pass strings as parameters or function results. This
  applies to all strings passed to and from your DLL--even those that
  are nested in records and classes. ShareMem is the interface unit to
  the BORLNDMM.DLL shared memory manager, which must be deployed along
  with your DLL. To avoid using BORLNDMM.DLL, pass string information
  using PChar or ShortString parameters. }

  // CodeCR424 LMS 2013-05-07
  //   Modified the DLL to farm out it's work into threads

  // LMS NOTE: System.SimpleShareMem,
  // Could have used the new SimpleShareMem or older ShareMem but decided on not using it
  //   as it could impact CAPRI in untested ways. So manually we allocate memory for the return
  //   type and send it from CAPRI.
  // Also switched to using cdecl instead of stdcall.

uses
  System.SysUtils,
  System.Classes,
  System.SyncObjs,
  System.DateUtils,
  Types,
  IOUtils,
  Ole2,
  SQLExpr,
  InvokeRegistry,
  SoapHTTPClient,
  System.TypInfo,
  Vcl.Dialogs,
  Vcl.Forms,
  Vcl.extctrls,
  Math,
  vva in 'vva.pas',
  WSSE in 'WSSE.pas';

type
  TCallBackFuncToGetDocBuffer = function(NewLength: Integer): Pointer of object; cdecl;

  TEventHandlers = class  //dummy class to allow use of event handler
  public
    procedure HTTPRIO1BeforeExecute(const MethodName: string; SOAPRequest: TStream);
  end;

  TThreadProcecdure = Procedure of object;

  TWaitThread = class( TThread )
  private
    FResultReady : Boolean;
    FResultFetched : Boolean;
    FExceptionRaised: Exception;
    ThreadTermTimer : TTimer;
    FStart : TDatetime;
    Procedure TimeOutThreadTerm(Sender: TObject);
    function getIsTerminated: boolean;
  protected
    procedure Execute; override;
  public
    RetGetDocumentThreadedTest : AnsiString;
    RetGetDocumentThreaded : PAnsiChar;
    _DocID, _DocSource, _DocFormat, _DocJRO,
      _Pth, _UserID, _URL, _SECUSR, _SECPWD, _SECTOKEN: WideString;
    ProcedureCalled : TThreadProcecdure;

    Property IsTerminated : Boolean read getIsTerminated;  // Make it not protected.

    property ResultReady: Boolean read FResultReady write FResultReady;
    property ResultFetched: Boolean read FResultFetched write FResultFetched;

    procedure DoGetDocumentThreaded;
    procedure DoGetDocumentThreadedTest;
    constructor Create(const Owner: TComponent; const CreateSuspended: Boolean); overload;
    destructor Destroy; override;
    procedure OnTerminate(Sender: TObject);
  end;

//var
  { ReturnedDocumentList must be a global variable so it's preserved in memory
    after the GetDocumentList function exits.
    LMS: Can't do that. There is no guarantee that the DLL is loaded and not
    ever going to be called from another instance. }
  // ReturnedDocumentList: AnsiString;

{$R *.res}

procedure DoProcessMessages(AppProxy: TObject; AppProxyProcessMessagesPointer : Pointer);
type
  TExec = procedure of object;
var
  Routine: TMethod;
  Exec: TExec;
begin
  Routine.Data := Pointer(AppProxy);
  Routine.Code := AppProxyProcessMessagesPointer; //AppProxy.MethodAddress('ProcessMessages');
  if Assigned(Routine.Code) then begin
    Exec := TExec(Routine);
    Exec;
  end;
end;

procedure WaitForThread(const AApp: TObject; const AppProcessMessages : Pointer; FWaitThread: TWaitThread);
begin
  repeat
    Sleep(200);
    DoProcessMessages(AApp, AppProcessMessages);
  until ((FWaitThread.ResultReady) or (FWaitThread.IsTerminated));
  if (FWaitThread.FExceptionRaised <> nil) then
    raise Exception.Create(FWaitThread.FExceptionRaised.Message);   // TODO: LMS May add more info here.
end;

{=============================================================================
 FileToByteArray
 This function loads the reads the bytes from the PDF file into a
 TByteDynArray type.
 =============================================================================}
function FileToByteArray(const FileName: string): TByteDynArray;
const BLOCK_SIZE=1024;
var BytesRead, BytesToWrite, Count : integer;
  F : File of Byte;
  pTemp : Pointer;
begin
  AssignFile( F, FileName );
  Reset(F);
  try
    Count := FileSize( F );
    SetLength(Result, Count );
    pTemp := @Result[0];
    BytesRead := BLOCK_SIZE;
    while (BytesRead = BLOCK_SIZE ) do
    begin
      BytesToWrite := Min(Count, BLOCK_SIZE);
      BlockRead(F, pTemp^, BytesToWrite , BytesRead );
      pTemp := Pointer(LongInt(pTemp) + BLOCK_SIZE);
      Count := Count-BytesRead;
    end;
  finally
    CloseFile( F );
  end;
end;

{ TEventHandlers }

{==========================================================================
 This event handler is assigned dynamically to the dynamically created
 HTTPRIO object.  The handler replaces the default 'NS' namespace tags
 in the security header with the web security namespaces required by the
 VVA web service.  Usually, the actual value of the namespace doesn't matter,
 but the VVA service would not work correctly with the Delphi assigned
 namespaces.
 ==========================================================================}
procedure TEventHandlers.HTTPRIO1BeforeExecute(const MethodName: string;
  SOAPRequest: TStream);
var
  aRequestString: TStringList;
begin
  aRequestString := TStringlist.Create;
  try
    SOAPRequest.Position := 0;
    aRequestString.LoadFromStream(SOAPRequest);
    aRequestString.Text := StringReplace(aRequestString.Text, ':NS1', ':wsse', [rfReplaceAll]);
    aRequestString.Text := StringReplace(aRequestString.Text, ':NS2', ':wsu', [rfReplaceAll]);
    aRequestString.Text := StringReplace(aRequestString.Text, 'NS1:', 'wsse:', [rfReplaceAll]);
    aRequestString.Text := StringReplace(aRequestString.Text, 'NS2:', 'wsu:', [rfReplaceAll]);
    aRequestString.Text := StringReplace(aRequestString.Text, '<Username', '<wsse:Username',[]);
    aRequestString.Text := StringReplace(aRequestString.Text, '</Username>', '</wsse:Username>',[]);
    SOAPRequest.Position := 0;
    aRequestString.SaveToStream(SOAPRequest);
  finally
    FreeAndNil(aRequestString);
  end;
end;

{ TWaitThread }

procedure TWaitThread.TimeOutThreadTerm(Sender: TObject);
const MAX_THREAD_LIFE_MINUTES = 45;
begin
  If (FResultFetched or (IncMinute(FStart, MAX_THREAD_LIFE_MINUTES) > Now)) then
    Terminate;
end;

procedure TWaitThread.Execute;
begin
  FResultReady := false;
  FResultFetched := false;

  ProcedureCalled;

  Self.Priority := TThreadPriority.tpIdle;
  repeat
    Self.Yield;
  until FResultReady or IsTerminated ;
  Self.Yield;

  ThreadTermTimer.Enabled := false;
  ThreadTermTimer.Interval := 2000; // 400 seconds for debugging.. otherwise Check every 2 seconds // Does not need to be high to finish even very large transfers.
  if (DebugHook <> 0) then
    ThreadTermTimer.Interval := 400000;

  ThreadTermTimer.Enabled := True;
end;

function TWaitThread.getIsTerminated: boolean;
begin
  Result := Terminated;
end;

procedure TWaitThread.OnTerminate(Sender: TObject);
begin
  (*
    Note: The method assigned to the OnTerminate event is executed in the context of the main thread rather
     than the context of the thread being terminated. This means you can access the user interface of your
     application safely without calling the Synchronize method.
  *)
  FreeAndNil(Self);
end;

destructor TWaitThread.Destroy;
begin
  ThreadTermTimer.Enabled := false;
  FreeAndNil(ThreadTermTimer);
  inherited;
end;

procedure TWaitThread.DoGetDocumentThreaded;
var
  aDocOperation: DocOperations;  //Proxy object
  Hdr: Security;   //Web service security header
  anEventHandler: TEventHandlers;  //Dummy event handler
  anHTTPRIO: THTTPRIO;  //Use separate Remote Interface Object so we can intercept events
  aGetDocContentResponse: GetDocumentContentResponse;
  aDocumentContent: DocumentContent;
  tStr : AnsiString;
const
  PWD_TYPE = 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText';
begin
  aDocumentContent := DocumentContent.Create;
  Hdr := Security.Create;
  anEventHandler := TEventHandlers.Create;
  {Use nil as owner to allow anHTTPRIO to manage its lifetime using the
   reference count}
  anHTTPRIO := THTTPRIO.Create(nil);
  anHTTPRIO.OnBeforeExecute := anEventHandler.HTTPRIO1BeforeExecute;
  try
    {Populate security header}
    Hdr.MustUnderstand := True;
    Hdr.UsernameToken := UsernameToken.Create;
    Hdr.UsernameToken.Id := _SECTOKEN;
    Hdr.UsernameToken.Username := _SECUSR;
    Hdr.UsernameToken.Password := PasswordString.Create;
    Hdr.UsernameToken.Password.Type_ := PWD_TYPE;
    Hdr.UsernameToken.Password.Text := _SECPWD;

    {Populate documentlist object}
    aDocumentContent.fnDcmntId := _DocID;
    aDocumentContent.fnDcmntSource := fnDcmntSource(GetEnumValue(TypeInfo(fnDcmntSource), _DocSource));
    aDocumentContent.dcmntFormatCd := dcmntFormatCd(GetEnumValue(TypeInfo(dcmntFormatCd), _DocFormat));
    aDocumentContent.jro := _DocJRO;
    aDocumentContent.userId := _UserID;
    try
      CoInitialize(nil);
      aDocOperation := GetDocOperations(False, _URL, anHTTPRIO);
      (aDocOperation as ISoapHeaders).Send(Hdr);
      aGetDocContentResponse := aDocOperation.GetDocumentContent(aDocumentContent);
      TFile.WriteAllBytes(_Pth, TBytes(aGetDocContentResponse.content));

      // LMS GetDocument := PAnsiChar(AnsiString('OK'));
      //      Mime Type is used for actual naming of file. docFormatCd is not the actual format of the file returned by webservice.
      tStr := AnsiString('^' + aGetDocContentResponse.mimeType);
      RetGetDocumentThreaded := PAnsiChar(AnsiString('OK')+tStr);
    except
      on E:Exception do
      begin
        RetGetDocumentThreaded := PAnsiChar(AnsiString('ERROR: ' + E.Message));
        FExceptionRaised := E;
      end;
    end;
    FResultReady := True;
  finally
    anHTTPRIO.OnBeforeExecute := nil;
    FreeAndNil(anEventHandler);
    aDocOperation := nil;
    FreeAndNil(Hdr);
    FreeAndNil(aDocumentContent);
  end;
end;

constructor TWaitThread.Create(const Owner: TComponent; const CreateSuspended: Boolean);
begin
  FreeOnTerminate := False; // True;
  FStart := now;
  FExceptionRaised := nil;
  FResultReady := false;
  FResultFetched := false;

  ThreadTermTimer := TTimer.Create(nil);
  ThreadTermTimer.Enabled := false;
  ThreadTermTimer.OnTimer := TimeOutThreadTerm;

  ThreadTermTimer.Interval := 2000000; // looong delay (33 mins), only for thread to cleanly terminate itself in case of exceptions where the Execute call doesn't get to enable the timer to terminate self.
  ThreadTermTimer.Enabled := true; //  If there is any exception, the looong delay timer will still cleanly terminate this thread

  inherited Create(CreateSuspended);
end;

Procedure TWaitThread.DoGetDocumentThreadedTest;
begin
  Try
    RetGetDocumentThreadedTest := 'Test Value Returned';
    FResultReady := True;
  except on e: exception do
    begin
      FExceptionRaised := E;
    end
  end;
end;

function GetDocumentTest(const _DocID: WideString; const AApp: TObject
                        ; const AppProcessMessages : Pointer
                        ): AnsiString; cdecl;
var
  FWaitThread : TWaitThread; // LMS Private variables to handle threading of SOAP calls.
Begin
  FWaitThread := TWaitThread.Create(nil, True) ;
  try
    FWaitThread._DocID := _DocID;
    FWaitThread.ProcedureCalled := FWaitThread.DoGetDocumentThreadedTest;
    FWaitThread.Start;
    WaitForThread(AApp, AppProcessMessages, FWaitThread);

    if FWaitThread.ResultReady then begin
      Result := FWaitThread.RetGetDocumentThreadedTest;
      FWaitThread.ResultFetched := True;
    end;
  finally
    FWaitThread.ThreadTermTimer.Enabled := true;  // ensure that the thread cleans itself
    FWaitThread.Terminate;
  end;
end;

{==============================================================================
 SendVVADoc
 This exported function is the web services client for the Virtual VA
 web service.
 Input:
   _ClmNmbr  - Veteran Claim Number
   _Pth      - PDF file path
   _DocTitle - document title
   _FilNme   - PDF file name
   _Doc Type - document type
   _DocSubj  - document subject
   _DocCat   - document category
   _Src      - document source
   _Usr      - person inserting document
   _SrcCmnt  - document source comment
   _URL      - web service URL
   _SECUSR   - web service security header user
   _SECPWD   - web service security header password
   _SECTOKEN - web service security header user token

 Output:  returns the results from the web service on success; otherwise,
          returns the contents of the exception text.
 ==============================================================================}
function SendVVADoc(const _ClmNmbr: WideString; const _Pth: WideString;
                    const _DocTitle: WideString; const _FilNme: WideString;
                    const _DocType: Integer; const _DocSubj: WideString;
                    const _DocCat: Integer; const _Src: WideString;
                    const _Usr: WideString; const _SrcCmnt: WideString;
                    const _URL: WideString;
                    const _SECUSR: WideString;
                    const _SECPWD: WideString;
                    const _SECTOKEN: WideString): ShortString; cdecl;
var
  aDocOperation: DocOperations;  //Proxy object
  doc: vaDocument;  //Document object
  docResponse: Int64;  //Int64
  Hdr: Security;   //Web service security header
  anEventHandler: TEventHandlers;  //Dummy event handler
  anHTTPRIO: THTTPRIO;  //Use separate Remote Interface Object so we can intercept events
const
  PWD_TYPE = 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText';
begin
  Result := '';
  doc := vaDocument.Create;
  Hdr := Security.Create;
  anEventHandler := TEventHandlers.Create;
  {Use nil as owner to allow anHTTPRIO to manage its lifetime using the
   reference count}
  anHTTPRIO := THTTPRIO.Create(nil);
  anHTTPRIO.OnBeforeExecute := anEventHandler.HTTPRIO1BeforeExecute;
  try
    {Populate security header}
    Hdr.MustUnderstand := True;
    Hdr.UsernameToken := UsernameToken.Create;
    Hdr.UsernameToken.Id := _SECTOKEN;
    Hdr.UsernameToken.Username := _SECUSR;
    Hdr.UsernameToken.Password := PasswordString.Create;
    Hdr.UsernameToken.Password.Type_ := PWD_TYPE;
    Hdr.UsernameToken.Password.Text := _SECPWD;

    {Populate document object}
    doc.content := FileToByteArray(_Pth);
    doc.documentTitle := _DocTitle;
    doc.fileName := _FilNme;
    doc.claimNumber := _ClmNmbr;
    doc.documentType := _DocType;
    doc.documentSubject := _DocSubj;
    doc.insertedByUserId := _Usr;
    doc.sourceComment := _SrcCmnt;
    doc.source := _Src;
    doc.documentCategory := _DocCat;
    try
      CoInitialize(nil);
      aDocOperation := GetDocOperations(False, _URL, anHTTPRIO);
      (aDocOperation as ISoapHeaders).Send(Hdr);
      docResponse := aDocOperation.addDocument(doc);
      Result := ShortString(IntToStr(docResponse));
    except
      on E:Exception do
      begin
        Result := ShortString(E.Message);
      end;
    end;
  finally
    anHTTPRIO.OnBeforeExecute := nil;
    FreeAndNil(anEventHandler);
    aDocOperation := nil;
    FreeAndNil(doc);
    FreeAndNil(Hdr);
  end;
end;

{==============================================================================
 GetDocumentList
 This exported procedure is the web services client for the Virtual VA
 web service.  It retrieves a list of documents available for retrieval.
 Input:
   _ClmNmbr  - Veteran Claim Number
   _URL      - web service URL
   _SECUSR   - web service security header user
   _SECPWD   - web service security header password
   _SECTOKEN - web service security header user token

 Output:  Returns the list of available documents as TStringList.Text with each
          entry containing caret-delimited document traits on success.
          Returns 'ERROR:' and the exception message text on failure

       Format:  Piece 1:  Date of Receipt  (rcvdDt)
                Piece 2:  DocTypeDesc   (dcmntTypeDescpTxt)
                Piece 3:  Document Subject (subjctTxt)
                Piece 4:  FileNet Document ID (fnDcmntId)
                Piece 5:  Document Format  (dcmntFormatCd)
                Piece 6:  Jurisdictional RO  (jrsdtnRoNbr)
                Piece 7:  Source Location (fnDcmntSource)
                Piece 8:  Restricted  (rstrcdDcmntInd)
                Piece 9:  Restricted Reason (rstrcdReasonLupId)
                Piece 10: Source Comment (sourceComntTxt)
                Piece 11: Document Record ID (dcmnRecordID)                      // JRL 3/1/13

  NOTE: PAnsiChar is used for the return value to provide compatibility
        betweem the unicode strings used in the dll from Delphi XE2 and the
        ansi strings used in the application from Delphi 2006.  Using
        PAnsiChar removes the need to use ShareMem and eliminates the need
        to distribute the BorlandMM.dll with the application.
        LMS: However using PAnsiChar adds the potential of memory corruption
        and is known to be inherently unsafe. Changing the function to Procedure
        that gets the pointer to an already pre-allocated memory buffer.
 ==============================================================================}
function GetDocumentList(const _ClmNmbr: WideString;
                         const _URL: WideString;
                         const _SECUSR: WideString;
                         const _SECPWD: WideString;
                         const _SECTOKEN: WideString;
                         const _SetDocListLength: TCallBackFuncToGetDocBuffer
                         ; const AApp: TObject
                         ; const AppProcessMessages : Pointer
                         ): PAnsiChar; cdecl;
var
  aDocOperation: DocOperations;  //Proxy object
  Hdr: Security;   //Web service security header
  anEventHandler: TEventHandlers;  //Dummy event handler
  anHTTPRIO: THTTPRIO;  //Use separate Remote Interface Object so we can intercept events
  aGetDocListResponse: GetDocumentListResponse;
  aDocumentList: DocumentList;
  DocList: TStringList;
  I: Integer;
  aSubject: String;
  ReturnedDocumentList: WideString;  // LMS Not using the DLL wide scoped vairable. Moved here.
  TempRetDocListAddr: Pointer;
  RetDocLen : Integer;
const
  PWD_TYPE = 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText';
begin
  Result := nil;
  ReturnedDocumentList := '';
  aDocumentList := DocumentList.Create;  //document object passed to the web svc
  DocList := TStringList.Create;  //document list results container
  Hdr := Security.Create;  //WSSE security header
  anEventHandler := TEventHandlers.Create;  //event handler the correct header

  {Use nil as owner to allow anHTTPRIO to manage its lifetime using the
   reference count}
  anHTTPRIO := THTTPRIO.Create(nil);
  anHTTPRIO.OnBeforeExecute := anEventHandler.HTTPRIO1BeforeExecute;
  try
    {Populate security header}
    Hdr.MustUnderstand := True;
    Hdr.UsernameToken := UsernameToken.Create;
    Hdr.UsernameToken.Id := _SECTOKEN;
    Hdr.UsernameToken.Username := _SECUSR;
    Hdr.UsernameToken.Password := PasswordString.Create;
    Hdr.UsernameToken.Password.Type_ := PWD_TYPE;
    Hdr.UsernameToken.Password.Text := _SECPWD;

    {Populate documentlist object used as web svc input parameter}
    aDocumentList.claimNbr := _ClmNmbr;
    try
      CoInitialize(nil);
      aDocOperation := GetDocOperations(False, _URL, anHTTPRIO);
      (aDocOperation as ISoapHeaders).Send(Hdr);
      aGetDocListResponse := aDocOperation.GetDocumentList(aDocumentList);
      if (aGetDocListResponse.size > 0) then  //populate returned results container
      begin
        for I := 0 to (aGetDocListResponse.size - 1) do
        begin
          {replace any line feeds or carets found in the subject with a space}
          aSubject := StringReplace(aGetDocListResponse.dcmntRecord[I].subjctTxt, #$A, ' ', [rfReplaceAll]);
          aSubject := StringReplace(aSubject, '^', ' ', [rfReplaceAll]);
          DocList.Add(aGetDocListResponse.dcmntRecord[I].rcvdDt + '^'
                    + aGetDocListResponse.dcmntRecord[I].dcmntTypeDescpTxt + '^'
                    + aSubject + '^'
                    + aGetDocListResponse.dcmntRecord[I].fnDcmntId + '^'
                    + aGetDocListResponse.dcmntRecord[I].dcmntFormatCd + '^'
                    + aGetDocListResponse.dcmntRecord[I].jrsdtnRoNbr + '^'
                    + aGetDocListResponse.dcmntRecord[I].fnDcmntSource + '^'
                    + aGetDocListResponse.dcmntRecord[I].rstrcdDcmntInd + '^'
                    + aGetDocListResponse.dcmntRecord[I].rstrcdReasonLupId + '^'
                    + aGetDocListResponse.dcmntRecord[I].sourceComntTxt + '^'
                    + aGetDocListResponse.dcmntRecord[I].dcmntRecordId);        // JRL 3/1/13
        end;

        ReturnedDocumentList := WideString(DocList.Text);
      end;
    except
      on E:Exception do
      begin
        ReturnedDocumentList := WideString('ERROR: ' + E.Message);
      end;
    end;
  finally  //populate results and cleanup objects
    anHTTPRIO.OnBeforeExecute := nil;
    FreeAndNil(anEventHandler);
    FreeAndNil(DocList);
    aDocOperation := nil;
    FreeAndNil(aDocumentList);
    FreeAndNil(Hdr);

    // Use _SetDocListLength Proc to set sizeof receiving buffer and also get the address of the
    // new buffer.
    // copy bytes from ReturnedDocumentList to _RetDocListPointer
    // Result := PAnsiChar(ReturnedDocumentList);

    if assigned(_SetDocListLength) then begin
      try
        if (length(ReturnedDocumentList) > 0) then begin
          RetDocLen := (Length(ReturnedDocumentList) +1) * SizeOf(WideChar);
          TempRetDocListAddr := (TCallBackFuncToGetDocBuffer(_SetDocListLength)(RetDocLen));
          System.Move(ReturnedDocumentList[1], TempRetDocListAddr^, RetDocLen);
          GetDocumentList := TempRetDocListAddr;   // Validation for caller that no exception occured.
        end;
      except
        on E:Exception do begin
          // no exceptions allowed to bubble.
          GetDocumentList := nil;
        end;
      end;
    end;
  end;
end;

{==============================================================================
 GetDocument
 This exported function is the web services client for the Virtual VA
 web service.  It retrieves the content of a selected document.
 Input:
   _DocID     - Document ID
   _DocSource - Document Source
   _DocFormat - Document Format
   _DocJRO    - Jurisdictional Regional Office
   _Pth       - Retrieved file destination path
   _UserID    - Person retrieving document
   _URL       - web service URL
   _SECUSR    - web service security header user
   _SECPWD    - web service security header password
   _SECTOKEN  - web service security header user token

 Output:  Returns 'OK' on success; Otherwise, returns 'ERROR:' and the exception
          message.
 ==============================================================================}
function GetDocument(const _DocID: WideString;
                      const _DocSource: WideString;
                      const _DocFormat: WideString;
                      const _DocJRO: WideString;
                      const _Pth: WideString;
                      const _UserID: WideString;
                      const _URL: WideString;
                      const _SECUSR: WideString;
                      const _SECPWD: WideString;
                      const _SECTOKEN: WideString
                      ; const AApp: TObject
                      ; const AppProcessMessages : Pointer
                      ): PAnsiChar; cdecl;
var
  FWaitThread : TWaitThread;
Begin
  Result := nil;
  FWaitThread := TWaitThread.Create(nil, True) ;
  try
    FWaitThread._DocID := _DocID;
    FWaitThread._DocSource := _DocSource;
    FWaitThread._DocFormat := _DocFormat;
    FWaitThread._DocJRO := _DocJRO;
    FWaitThread._Pth := _Pth;
    FWaitThread._UserID := _UserID;
    FWaitThread._URL := _URL;
    FWaitThread._SECUSR := _SECUSR;
    FWaitThread._SECPWD := _SECPWD;
    FWaitThread._SECTOKEN := _SECTOKEN;
    FWaitThread.ProcedureCalled := FWaitThread.DoGetDocumentThreaded;
    FWaitThread.Start;
    WaitForThread(AApp, AppProcessMessages, FWaitThread);

    if FWaitThread.ResultReady then begin
      Result := FWaitThread.RetGetDocumentThreaded;
      FWaitThread.ResultFetched := True;
    end;
  finally
    FWaitThread.ThreadTermTimer.Enabled := true;  // ensure that the thread cleans itself
    FWaitThread.Terminate;
  end;
End;

exports
  SendVVADoc, GetDocumentList, GetDocument, GetDocumentTest;
end.
